==> ciao-4.6-src-core.tar.gz: ciao-4.6/src/lib/help/LookupTable.cc <==
// 
//  Copyright (C) 2002,2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "LookupTable.hh"


// Needed for hashing function

#define mix64(a,b,c) \
{ \
  a -= b; a -= c; a ^= (c>>43); \
  b -= c; b -= a; b ^= (a<<9); \
  c -= a; c -= b; c ^= (b>>8); \
  a -= b; a -= c; a ^= (c>>38); \
  b -= c; b -= a; b ^= (a<<23); \
  c -= a; c -= b; c ^= (b>>5); \
  a -= b; a -= c; a ^= (c>>35); \
  b -= c; b -= a; b ^= (a<<49); \
  c -= a; c -= b; c ^= (b>>11); \
  a -= b; a -= c; a ^= (c>>12); \
  b -= c; b -= a; b ^= (a<<18); \
  c -= a; c -= b; c ^= (b>>22); \
}


/*
**
** Desc: Constructor
**
**
** Pre: inSize size of lookup table
**
** Post: Object is allocated and initialized
**       
**
**
*/


LookupTable::LookupTable(unsigned long long  inSize)
{ 
  table = new Bucket*[inSize];
  for(int indx=0;indx<inSize;indx++) 
    {
      table[indx]=0;
    };
  size=inSize;

};

/*
**
** Desc: Destructor
**
**
** Pre: None
**      
**      
**
** Post: Dealocation
**       
**
**
*/


LookupTable::~LookupTable()
{

  for(int indx=0;indx<size;indx++)
    {
      delete table[indx];
    }
  delete[] table;

};




/*
**
** Desc: hashes string of chars to  long long values
**
**
** Pre: k input string
**      length length of k
**      level init value for hashing
**
** Post: string pointed to by k of length length is hashed into long long value that 
**       is returned by this function
**       
**
**
*/

unsigned long long LookupTable::Hash( register char *k, register unsigned long long length, register unsigned long long  level)
{
  register unsigned long long a,b,c,len;

  // Set up the internal state 
  len = length;
  a = b = level;                         // the previous hash value 
  c = 0x9e3779b97f4a7c13LL; // the golden ratio; an arbitrary value 

  //---------------------------------------- handle most of the key 
  while (len >= 24)
  {
    a += (k[0]        +((unsigned long long)k[ 1]<< 8)+((unsigned long long)k[ 2]<<16)+((unsigned long long)k[ 3]<<24)
     +((unsigned long long)k[4 ]<<32)+((unsigned long long)k[ 5]<<40)+((unsigned long long)k[ 6]<<48)+((unsigned long long)k[ 7]<<56));
    b += (k[8]        +((unsigned long long)k[ 9]<< 8)+((unsigned long long)k[10]<<16)+((unsigned long long)k[11]<<24)
     +((unsigned long long)k[12]<<32)+((unsigned long long)k[13]<<40)+((unsigned long long)k[14]<<48)+((unsigned long long)k[15]<<56));
    c += (k[16]       +((unsigned long long)k[17]<< 8)+((unsigned long long)k[18]<<16)+((unsigned long long)k[19]<<24)
     +((unsigned long long)k[20]<<32)+((unsigned long long)k[21]<<40)+((unsigned long long)k[22]<<48)+((unsigned long long)k[23]<<56));
    mix64(a,b,c);
    k += 24; len -= 24;
  }

  //------------------------------------- handle the last 23 bytes 
  c += length;
  switch(len)              // all the case statements fall through
  {
  case 23: c+=((unsigned long long)k[22]<<56);
  case 22: c+=((unsigned long long)k[21]<<48);
  case 21: c+=((unsigned long long)k[20]<<40);
  case 20: c+=((unsigned long long)k[19]<<32);
  case 19: c+=((unsigned long long)k[18]<<24);
  case 18: c+=((unsigned long long)k[17]<<16);
  case 17: c+=((unsigned long long)k[16]<<8);
    /* the first byte of c is reserved for the length */
  case 16: b+=((unsigned long long)k[15]<<56);
  case 15: b+=((unsigned long long)k[14]<<48);
  case 14: b+=((unsigned long long)k[13]<<40);
  case 13: b+=((unsigned long long)k[12]<<32);
  case 12: b+=((unsigned long long)k[11]<<24);
  case 11: b+=((unsigned long long)k[10]<<16);
  case 10: b+=((unsigned long long)k[ 9]<<8);
  case  9: b+=((unsigned long long)k[ 8]);
  case  8: a+=((unsigned long long)k[ 7]<<56);
  case  7: a+=((unsigned long long)k[ 6]<<48);
  case  6: a+=((unsigned long long)k[ 5]<<40);
  case  5: a+=((unsigned long long)k[ 4]<<32);
  case  4: a+=((unsigned long long)k[ 3]<<24);
  case  3: a+=((unsigned long long)k[ 2]<<16);
  case  2: a+=((unsigned long long)k[ 1]<<8);
  case  1: a+=((unsigned long long)k[ 0]);
    /* case 0: nothing left to add */
  }
  mix64(a,b,c);
  //-------------------------------------------- report the result 
  return c;
};


/*
**
** Desc: overloaded indexing operator
**
**
** Pre: inStringIndex index to be accessed
**      
**      
**
** Post: returns int value stored at index inStringsIndex
**       
**       
**
**
*/

int& LookupTable::operator [](const AhelpStrings& inStringsIndex)
{
  Bucket* tmpPtr;
  unsigned long long indx=Hash(inStringsIndex.chars(),inStringsIndex.length(),0) % size;


  if ( table[indx]== 0 )
    {
      table[indx] = new Bucket();
      table[indx]->index=inStringsIndex;
      return table[indx]->value;
    }
  else
    {
      tmpPtr=table[indx];
      while(1)
	{
	  
	  if (tmpPtr->index == inStringsIndex )
	    {
	      break;
	    }
	  else if ( tmpPtr->next != 0 )
	    {
	      tmpPtr=tmpPtr->next;
	    }
	  else
	    {
	      tmpPtr->next = new Bucket();
	      tmpPtr->index = inStringsIndex;
	      tmpPtr = tmpPtr->next;
	      break;
	    };
	};
    };

  return tmpPtr->value;

};

==> ciao-4.8-src-core.tar.gz: ciao-4.8/src/lib/help/LookupTable.cc <==
// 
//  Copyright (C) 2002,2007,2015  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "LookupTable.hh"


// Needed for hashing function

#define mix64(a,b,c) \
{ \
  a -= b; a -= c; a ^= (c>>43); \
  b -= c; b -= a; b ^= (a<<9); \
  c -= a; c -= b; c ^= (b>>8); \
  a -= b; a -= c; a ^= (c>>38); \
  b -= c; b -= a; b ^= (a<<23); \
  c -= a; c -= b; c ^= (b>>5); \
  a -= b; a -= c; a ^= (c>>35); \
  b -= c; b -= a; b ^= (a<<49); \
  c -= a; c -= b; c ^= (b>>11); \
  a -= b; a -= c; a ^= (c>>12); \
  b -= c; b -= a; b ^= (a<<18); \
  c -= a; c -= b; c ^= (b>>22); \
}


/*
**
** Desc: Constructor
**
**
** Pre: inSize size of lookup table
**
** Post: Object is allocated and initialized
**       
**
**
*/


LookupTable::LookupTable(unsigned long long  inSize)
{ 
  table = new Bucket*[inSize];
  for(unsigned long long indx=0;indx<inSize;indx++) 
    {
      table[indx]=0;
    };
  size=inSize;

};

/*
**
** Desc: Destructor
**
**
** Pre: None
**      
**      
**
** Post: Dealocation
**       
**
**
*/


LookupTable::~LookupTable()
{

  for(unsigned long long indx=0;indx<size;indx++)
    {
      delete table[indx];
    }
  delete[] table;

};




/*
**
** Desc: hashes string of chars to  long long values
**
**
** Pre: k input string
**      length length of k
**      level init value for hashing
**
** Post: string pointed to by k of length length is hashed into long long value that 
**       is returned by this function
**       
**
**
*/

unsigned long long LookupTable::Hash( register char *k, register unsigned long long length, register unsigned long long  level)
{
  register unsigned long long a,b,c,len;

  // Set up the internal state 
  len = length;
  a = b = level;                         // the previous hash value 
  c = 0x9e3779b97f4a7c13LL; // the golden ratio; an arbitrary value 

  //---------------------------------------- handle most of the key 
  while (len >= 24)
  {
    a += (k[0]        +((unsigned long long)k[ 1]<< 8)+((unsigned long long)k[ 2]<<16)+((unsigned long long)k[ 3]<<24)
     +((unsigned long long)k[4 ]<<32)+((unsigned long long)k[ 5]<<40)+((unsigned long long)k[ 6]<<48)+((unsigned long long)k[ 7]<<56));
    b += (k[8]        +((unsigned long long)k[ 9]<< 8)+((unsigned long long)k[10]<<16)+((unsigned long long)k[11]<<24)
     +((unsigned long long)k[12]<<32)+((unsigned long long)k[13]<<40)+((unsigned long long)k[14]<<48)+((unsigned long long)k[15]<<56));
    c += (k[16]       +((unsigned long long)k[17]<< 8)+((unsigned long long)k[18]<<16)+((unsigned long long)k[19]<<24)
     +((unsigned long long)k[20]<<32)+((unsigned long long)k[21]<<40)+((unsigned long long)k[22]<<48)+((unsigned long long)k[23]<<56));
    mix64(a,b,c);
    k += 24; len -= 24;
  }

  //------------------------------------- handle the last 23 bytes 
  c += length;
  switch(len)              // all the case statements fall through
  {
  case 23: c+=((unsigned long long)k[22]<<56);
  case 22: c+=((unsigned long long)k[21]<<48);
  case 21: c+=((unsigned long long)k[20]<<40);
  case 20: c+=((unsigned long long)k[19]<<32);
  case 19: c+=((unsigned long long)k[18]<<24);
  case 18: c+=((unsigned long long)k[17]<<16);
  case 17: c+=((unsigned long long)k[16]<<8);
    /* the first byte of c is reserved for the length */
  case 16: b+=((unsigned long long)k[15]<<56);
  case 15: b+=((unsigned long long)k[14]<<48);
  case 14: b+=((unsigned long long)k[13]<<40);
  case 13: b+=((unsigned long long)k[12]<<32);
  case 12: b+=((unsigned long long)k[11]<<24);
  case 11: b+=((unsigned long long)k[10]<<16);
  case 10: b+=((unsigned long long)k[ 9]<<8);
  case  9: b+=((unsigned long long)k[ 8]);
  case  8: a+=((unsigned long long)k[ 7]<<56);
  case  7: a+=((unsigned long long)k[ 6]<<48);
  case  6: a+=((unsigned long long)k[ 5]<<40);
  case  5: a+=((unsigned long long)k[ 4]<<32);
  case  4: a+=((unsigned long long)k[ 3]<<24);
  case  3: a+=((unsigned long long)k[ 2]<<16);
  case  2: a+=((unsigned long long)k[ 1]<<8);
  case  1: a+=((unsigned long long)k[ 0]);
    /* case 0: nothing left to add */
  }
  mix64(a,b,c);
  //-------------------------------------------- report the result 
  return c;
};


/*
**
** Desc: overloaded indexing operator
**
**
** Pre: inStringIndex index to be accessed
**      
**      
**
** Post: returns int value stored at index inStringsIndex
**       
**       
**
**
*/

int& LookupTable::operator [](const AhelpStrings& inStringsIndex)
{
  Bucket* tmpPtr;
  unsigned long long indx=Hash(inStringsIndex.chars(),inStringsIndex.length(),0) % size;


  if ( table[indx]== 0 )
    {
      table[indx] = new Bucket();
      table[indx]->index=inStringsIndex;
      return table[indx]->value;
    }
  else
    {
      tmpPtr=table[indx];
      while(1)
	{
	  
	  if (tmpPtr->index == inStringsIndex )
	    {
	      break;
	    }
	  else if ( tmpPtr->next != 0 )
	    {
	      tmpPtr=tmpPtr->next;
	    }
	  else
	    {
	      tmpPtr->next = new Bucket();
	      tmpPtr->index = inStringsIndex;
	      tmpPtr = tmpPtr->next;
	      break;
	    };
	};
    };

  return tmpPtr->value;

};
